# Library modules
require 'abspath'
require 'ftools'

# My modules
require 'config'
require 'template'
require 'myxml'
require 'photo'
require 'urlencode'
require 'rp_debug'

class AlbumReader < MyXMLParser

  attr_writer :album

  def init
    @current_text = ""
    @got_title = false
    @got_date  = false
    @got_order = false
  end

  def startElement(name, attrs)
  end

  def endElement(name)
    current_text = @current_text.strip

    case name
    when 'title'
      @album.title = current_text unless current_text.empty?
      @got_title = true
    when 'thumbnail'
      @album.thumbnail = current_text
    when 'shortdesc'
      @album.short_desc = current_text
    when 'longdesc'
      @album.long_desc = current_text
    when 'date'
      @album.date = current_text
      @got_date = true
    when 'order'
      @album.order = current_text
      @got_order = true
    end

    # Speed things up a bit if we can.  We could do even better
    # by looking at the sorting policy for the parent album, but
    # that's probably overkill.
# Calling the `done' method horribly breaks ruby, so we can't do this :-(
#    done if $tree_only && @got_title
#    done if $albums_only && @got_title && @got_date && @got_order

    @current_text = ""
  end

  def character(text)
    @current_text += text
  end

end

class Album

  include Debug

  attr_accessor :config, :title, :short_desc, :thumbnail, :date, :order
  attr_reader :albums, :photos, :directory, :parent, :default_photo_path,
              :total_pages, :current_page, :sorter, :root, :directory_name
  attr_writer :long_desc
  protected :albums

  def initialize(directory=".", parent=nil)
    @parent          = parent
    @parents         = nil
    @root            = parent ? parent.root : self
    @directory       = directory
    @directory_name  = directory
    @directory_path  = nil
    @name            = ""
    @title           = @directory.gsub(/_/, ' ')
    @date            = ""
    @order           = ""
    @thumbnail       = ""
    @short_desc      = ""
    @long_desc       = ""
    @current_page    = 1

    @photos = Array.new
    @albums = Array.new
    @photos_by_filename = Hash.new
    @albums_by_dir      = Hash.new

    read_config
    read_album_xml
    @sorter = sorter_proc
    read_directory
  end
  
  def self_link
    link = directory + '/'
    link += @config.get_option('index_file') \
      unless @config.get_option('drop_index') == 'yes'
    return link
  end

  def top_dir
    # Relative to within self
    '../' * (parents.length - 1)
  end

  def top_page_url
    # Relative to within self
    url = top_dir
    url += @config.get_option('index_file') \
      unless @config.get_option('drop_index') == 'yes'
    url
  end

  def read_album_xml
    if File.exist?(file_path)
      debug 1, "  parsing " + file_path
      parser = AlbumReader.new
      parser.album = self
      parser.parse(file_path)
    end
  end

  def parents
    @parents ||= calculate_parents
  end

  def calculate_parents
    return @parent.parents + [ self ] if @parent
    return [ self ]
  end

  private :calculate_parents

  def parent_path
    parents.map { |p| p.title }.join(' > ')
  end

  def read_directory
    ignorelist = @config.globbed_ignore_list(directory_path)

    photos    = Array.new
    subalbums = Array.new

    Dir.entries(directory_path).each { |fname|
      fname =~ /^.+\.(.*?)$/
      next if fname == "." || fname == ".."
      next if thumbnail?(fname) || ignorelist[fname]

      if test(?d, directory_path + '/' + fname)
        subalbums.push create_album(fname)
        next
      end
      
      next unless $1 =~ /(jpe?g|png|gif)$/i

      photos.push create_photo(fname)
    }

    subalbums.sort.each { |sub_album|
      if sub_album.albums.length > 0 || sub_album.photos.length > 0 ||
          sub_album.class.name == 'AlbumLink'
        add_album(sub_album)
      end
    }

    photos.sort.each { |photo| add_photo(photo) }
  end
  
  def <=>(other)
    @parent.sorter[self, other]
  end

  def sorter_proc
    sort_by = @config.get_option('sort_subalbums_by')

    descending = (sort_by.gsub!(/^reverse\s*/, '')) ? -1 : 1

    case sort_by
    when 'filename'
      proc { |x, y|
        (x.directory.downcase <=> y.directory.downcase) * descending
      }
    when 'mtime'
      proc { |x, y|
        a = File.new(x.directory_path)
        b = File.new(y.directory_path)
        (a.mtime <=> b.mtime) * descending
      }
    when 'title'
      proc { |x, y|
        (x.title.downcase <=> y.title.downcase) * descending
      }
    when 'date'
      proc { |x, y|
        (x.date <=> y.date) * descending
      }
    when 'order_field'
      proc { |x, y|
        if (x.order.empty? && y.order.empty?)
          0
        elsif x.order.empty?
          1
        elsif y.order.empty?
          -1
        else
          x.order.to_f <=> y.order.to_f
        end * descending
      }
    else
       "Invalid value for 'sort_subalbums_by' config option: '#{sort_by}'"
    end
  end

  private :sorter_proc

  def thumbnail?(filename)
    filename =~ /^.*_\d+x\d+\.(jpe?g|gif|png)$/i
  end

  def create_album(fname)
    fullpath = directory_path + '/' + fname
    if test(?l, fullpath) && link_path = link_abspath(fullpath)
      album = AlbumLink.new(fname, self, link_path)
    else 
      album = Album.new(fname, self)
    end
  end

  # Figure out whether link points outside image hierarchy by
  # comparing path prefixes.
  def link_abspath(fname)
    link = File.readlink(fname)
    toplevel_abspath  = File.abs_path($input_dir).split('/')
    link_abspath      = File.abs_path(fname).split('/')

    while toplevel_abspath[0] == link_abspath[0]
      toplevel_abspath.shift
      link_abspath.shift
    end

    return (toplevel_abspath.nitems == 0) ? link_abspath.join('/') : false
  end
    

  def create_photo(fname)
    photo = Photo.new(fname, directory_path, @config)

    xml_location = photo.xml_path
    return photo unless xml_location

    preader = PhotoReader.new
    preader.photo = photo
    preader.config = config
    preader.parse(xml_location)
    return photo
  end

  private :create_photo

  # This routine is broken because you can't override initialize in the XML
  # Parser class.
  def read_config
    if @parent
      @config = Config.new(@parent.config)
    else 
      @config = Config.new
      @config.parse($config_file)
    end
    
    # If we have a config.xml in this per-directory,
    # override settings from it.
    local_config = directory_path + "/config.xml"
    if File.exist?(local_config)
      @config.parse(local_config)
    end
  end

  private :read_config, :read_album_xml, :read_directory

  def add_photo(photo)
    photo.parent      = self
    photo.album_index = @photos.length # starts at 0
    @photos.push(photo)
    @photos_by_filename[photo.filename] = photo
  end

  def add_album(album)
    @albums.push(album)
    @albums_by_dir[album.directory_name] = album
  end

  def output_html
    output_album_html
    @albums.each do |album|
      album.output_html
    end
    debug 1, parent_path
  end

  def output_album_html
    template = AlbumTemplate.get(@config)

    photos_per_page = @config.get_option('table_columns').to_i *
                      @config.get_option('table_rows').to_i

    @total_pages = (@photos.length.to_f / photos_per_page).ceil
    @total_pages = 1 if @total_pages == 0
    bottom = 0
    @current_page = 1
    while bottom <= @photos.length
      page = @photos[bottom .. bottom+photos_per_page-1]

      filename = index_filename
      File.open(directory_path + "/" + filename, 'w') { |file|
        output_page(file, template, page)
      }
      bottom += photos_per_page
      @current_page += 1
    end
    @current_page = 1
  end

  def index_filename(page_number = nil)
    page_number ||= @current_page
    filename = @config.get_option('index_file')

    if page_number != 1
      filename =~ /^(.+)\.(.*?)$/
      filename = $1 + "-" + (page_number - 1).to_s + "." + $2
    end

    return filename
  end

  def output_page(file, template, thumbs)
    file.puts template.albumheader(self)
    output_subalbums(file, template)
    output_thumbs(file, template, thumbs)
    file.puts template.albumfooter(self)
  end

  def output_tree_html
    tree_template = TreeTemplate.get(@config)  
    File.open(directory_path + "/" + @config.get_option('tree_file'), 'w') { |file|
      file.puts tree_template.treeheader(self)
      file.puts tree_template.tree(self)
      file.puts tree_template.treefooter
    }
  end

  ## Strictly this should only be done if the subalbums_same_page
  ## option is set to 'true', but I'm not going to worry about
  ## that for now.
  def output_subalbums(file, template)
    return unless @albums.length > 0
    file.puts(template.beginindex)
    @albums.each do |album|
      file.puts(template.indexentry(album))
    end
    file.puts(template.endindex)
  end

  def output_thumbs(file, template, thumbs)
    return if @photos.length == 0

    num_columns = @config.get_option('table_columns').to_i
    bottom = 0

    file.puts template.beginthumbs(self)
    while bottom < thumbs.length
      output_row(file, template, thumbs[bottom..bottom+num_columns-1],
                 thumbs.length <= num_columns)
      bottom += num_columns
    end
    file.puts template.endthumbs
  end

  def output_row(file, template, row, only_row)
    num_columns = @config.get_option('table_columns').to_i
    file.puts %/<tr align="left" valign="bottom">/
    row.each do |photo|
      file.puts template.thumbnail(photo)
      photo.output_html unless $albums_only
    end
    if ! only_row
      if row.length < num_columns
        extra = num_columns - row.length
        extra.times do 
          file.puts template.blankentry
        end
      end
    end
    file.puts "</tr>"
  end

  def index_file_url(page_number = nil)
    link = index_filename(page_number)
    if @config.get_option('drop_index') == 'yes' &&
       link == @config.get_option('index_file')
       link = ''
    end
    return link
  end

  def tree_file_url
    back = "../" * (parents.length - 1)
    back + @config.get_option('tree_file')
  end

  def tree_file_link
    %|<a href="#{tree_file_url}">tree</a>|
  end

  def dots 
    ' <span class="treepathdots">::</span> ' 
  end

  def tree(link=false)
    tree = [ tree_file_link ]
    depth = parents.length - 1
    parents.each { |p| 
      tree.push p.tree_link(depth, link)
      depth -= 1
    }
    tree.join(dots)
  end


  def tree_link(depth, link)
    back = ""
    back = "../" * depth if depth > 0
    back += index_file_url
    back = '.' if back.empty?

    return '<a href="' + back + '">' + title + '</a>' if (depth > 0 || link)
    return title
  end

  protected :tree_link

  def tree_page_entry
    entry = "<li>" + link_to_self + "\n"
    return entry unless @albums.length > 0

    entry += "<ul>\n"
    @albums.each { |album|
      entry += album.tree_page_entry
    }
    entry += "</ul>\n"
    entry
  end

  def link_to_self
    url = url_encode(directory_path)
    "<a href=\"#{url}/#{index_file_url}\">#{title}</a>"
  end

  def top_url
    return nil unless parent
    top_page_url
  end

  def up_url
    return nil unless parent
    '../' + parent.index_file_url
  end
  
  def next_url
    return nil unless current_page < total_pages
    index_file_url(current_page+1)
  end

  def prev_url
    return nil unless current_page > 1
    link = index_file_url(current_page-1)
    link = '.' if link.empty?
    link
  end

  def first_url
    return nil if total_pages == 1 || current_page == 1
    link = index_file_url(1)
    link = '.' if link.empty?
    return link
  end

  def last_url
    return nil if total_pages == 1 || current_page == total_pages
    return index_file_url(total_pages)
  end

  def first_photo_url
    
  end

  def last_photo_url
  end

  def filename
    @config.get_option('album_file')
  end

  def file_path
    directory_path + "/" + filename  
  end

  def directory_path
    @directory_path ||= parents.map { |p| p.directory }.join('/')
  end

  def photo_count
    count = @photos.length
    @albums.each { |album| count += album.photo_count }
    return count
  end
    
  def photo_count_text
    count = photo_count
    return "" if count == 0
    return count.to_s + "&nbsp;photo" if count == 1
    return count.to_s + "&nbsp;photos" if count > 1
  end

  def album_count
    count = @albums.length
    @albums.each { |album| count += album.album_count }
    return count
  end
  
  def album_count_text
    count = album_count
    return "" if count == 0
    return count.to_s + "&nbsp;sub-album" if count == 1
    return count.to_s + "&nbsp;sub-albums" if count > 1
  end

  def count
    pcount = photo_count_text
    acount = album_count_text

    if pcount == "" && acount == ""
      return ""
    elsif acount == ""
      return pcount
    elsif pcount == ""
      return acount
    end

    [pcount, acount].join(',&nbsp;')
  end

  def long_desc
    return @long_desc if @long_desc != ""
    @short_desc
  end

  def default_photo
    @default_photo_path = directory

    # Deal with <thumbnail> foo/bar.jpg </thumbnail>
    if @thumbnail != ""
      photo, *paths = rec_find_photo_by_path(@thumbnail)
      if photo
        @default_photo_path += '/' + paths.join('/') unless paths.empty?
        return photo
      else
        # Maybe it's a subalbum?
        subpath = @thumbnail.chomp('/')
        subalbum = rec_find_album_by_path(subpath)
        if subalbum
          # Have to call default_photo first to ensure that the
          # default_photo_path gets set.
          default_photo = subalbum.default_photo
          @default_photo_path = directory + '/' + subalbum.default_photo_path
          return default_photo
        else
          die "Invalid <thumbnail> setting for album `#{title}': #@thumbnail"
        end
      end
    end

    # No default photo specified, take first one in list.
    return @photos[0] if @photos[0]

    # No photos in this directory, recurse until we find one
    @albums.each { |subalbum|
      if subalbum.default_photo
        @default_photo_path = directory + "/" + subalbum.default_photo_path 
        return subalbum.default_photo 
      end
    }
  end

  def rec_find_album_by_path(path)
    return @albums_by_dir[path] unless path =~ %r!/!

    path =~ %r!(.+?)/(.*)!
    subalbum_dir, subpath = $1, $2
    subalbum = @albums_by_dir[subalbum_dir]
    die "Album #{directory_path} had no such subalbum `#{subalbum_dir}'" \
      unless subalbum

    subalbum.rec_find_album_by_path(subpath)
  end

  def rec_find_photo_by_path(path)
    return @photos_by_filename[path] unless path =~ %r!/!

    path =~ %r!(.+?)/(.*)!
    subalbum_dir, subpath = $1, $2
    subalbum = @albums_by_dir[subalbum_dir]
    die "Default thumbnail for `#{title}' album specified non-existent sub-album `#{subalbum_dir}'!" unless subalbum

    photo, *subpaths = subalbum.rec_find_photo_by_path(subpath)
    subpaths = subalbum_dir, *subpaths if subalbum_dir != ''
    return photo, subpaths
  end

  def die(msg)
    debug 1, msg
    exit 1
  end
end

class AlbumLink < Album
  
  def initialize(fname, parent, link_path)
    # link_path must be set before super is called
    @link_path      = link_path
    super(fname, parent)
    @linked_album   = nil
    @title          = fname
    @directory_name = fname
    debug 2, "    * symlink -> #{directory}"
  end

  # Does nothing
  def read_directory
  end

  # Does nothing  
  def output_album_html
  end

  def linked_album
    return @linked_album if @linked_album
    
    @linked_album = @root.rec_find_album_by_path(@link_path)
    die "Failed to find album linked from #{directory_path} by path `#{link_path}'" \
      unless @linked_album
    return @linked_album
  end
    
  def up_to_top
    # Relative to self's containing album
    '../' * (parents.length - 2)
  end

  def directory
    @my_directory ||= up_to_top + @link_path
  end
  
  def parent_path
    super + '@'
  end

  def tree_page_entry
    "<li>" + link_to_self + "@\n"
  end

  def link_to_self
    url = url_encode(@link_path)
    "<a href=\"#{url}/#{index_file_url}\">#{title}</a>"
  end

  def count
    linked_album.count
  end

  def default_photo
    # Have to call default_photo first to ensure that the
    # default_photo_path gets set.
    default_photo = linked_album.default_photo
    @default_photo_path = directory.gsub(%r![^/]+$!, '') \
      + linked_album.default_photo_path
    return default_photo
  end
  
end
