require 'myxml'
require 'urlencode'


# Utility function to find the closest page to a particular image_category for
# a particular photo.
# Template information class

class TemplateParser < MyXMLParser

  attr_writer :template

  def init
    @cdata = ""
  end

  def parse(template, config, template_type)
    @template = template
    @config   = config

    template_file = template_type + '.xml'

    # If it's just a filename, stick the template_dir on the front
    template_dir = @config.get_option('template_dir')
    template_file = template_dir + '/' + template_file \
      unless template_file =~ %r!/!

    raise "File does not exist: #{template_file}" unless test(?e, template_file)

    super(template_file)
  end

  def startElement(name, attrs)
    @cdata = "" 
  end

  def endElement(name)
    @template.set_template(name, @cdata)
  end

  def character(text)
    @cdata += text
  end

end

class HTMLTemplate

  def initialize(config)
    @config    = config
    @templates = Hash.new

    parser = TemplateParser.new
    parser.parse(self, @config, 'shared')
  end

  @@cache = Hash.new

  def HTMLTemplate.get(config)
    return @@cache[self] if @@cache[self]

    @@cache[self] = self.new(config)
    @@cache[self].read
    return @@cache[self]
  end

  def read
    parser = TemplateParser.new
    parser.parse(self, @config, template_name)
  end

  def set_template(name, value)
    @templates[name] = value
  end

  def method_missing(symbol)
    return @templates[symbol.id2name].dup if @templates[symbol.id2name]
    raise "No such template " + symbol.id2name
  end 
  
  def image_links(photo, page_type=nil)
    @config.image_categories.map do |image_category|
      page_link(image_category, photo, page_type)
    end.delete_if { |link| link == "" }.join '&nbsp;&nbsp;'
  end
  
  def image_long_links(photo, page_type=nil)
    @config.image_categories.map do |image_category|
      page_link(image_category, photo, page_type, true)
    end.delete_if { |link| link == "" }.join '&nbsp;&nbsp;'
  end
  
  def page_link(image_category, photo, current_type, long=false)
    return "" unless photo.size_possible?(image_category)

    if current_type && image_category.type_name == current_type.type_name
      return image_category.label(long)
    end

    new_filename = url_encode(image_category.page_filename(photo.filename))
    return %!<a href="#{new_filename}">#{image_category.label(long)}</a>!
  end
  
  def make_substitutions(text, substs)
    text = subst_includes(text)
    
    substs.each do |name, value|
      if text =~ /\[#{name}\]/
        value = value.call if value.type.to_s == 'Proc'
        text.gsub!(/\[#{name}\]/, value)
      end
    end
    return text
  end

  def subst_includes(text)
    text.gsub(/\{(.+?)\}/) { |match|
      subst_includes(get_template($1))
    }
  end

  def get_template(name)
    if ! @templates.has_key?(name)
      raise "No such template '#{name}'"
    end
    @templates[name].dup
  end

  def find_closest_category(image_category, photo)
    index = @config.image_categories.index(image_category)

    while index >= 0
      next_type = @config.image_categories[index]
      return next_type if photo.size_possible?(next_type)
      index -= 1
    end
    
    return @config.image_categories[-1]  # default category always at end
  end
  
end

class AlbumTemplate < HTMLTemplate

  def template_name
    'album'
  end

  def album_substitutions(text, substs, album)
    text = make_substitutions(text, substs)

    # This is a bit ugly, since we have two layers of
    # substitutions folded into one.
    top_subbed = get_template('top').gsub(/\[TOP\]/, album.top_dir)
    up_subbed  = get_template('up').gsub(/\[TOP\]/,  album.top_dir)

    up_link = if album.parent
                '<a href="' + album.up_url + '" title="Up to parent">' \
                + up_subbed  + '</a>'
              else
                '&nbsp;'
              end

    top_link =
      album.parent ? \
        '<a href="' + album.top_page_url + '" title="Up to top">' + \
         top_subbed + '</a>' \
       : '&nbsp;'

    global_substs = {
      'TOP'       => proc { album.top_dir    },
      'UPLINK'    => proc { up_link          },
      'TOPLINK'   => proc { top_link         },
      'HEADLINKS' => proc { headlinks(album) },
    }

    return make_substitutions(text, global_substs)
  end

  def headlinks(album)
    headlinks = <<EOF
  <link rel="toc" href="#{album.tree_file_url}">
EOF

    %w|top start up first last prev next|.each { |dir|
      method = dir
      method = 'top' if method == 'start'
      url = album.send(method + '_url')
      headlinks += <<EOF if url
  <link rel="#{dir}" href="#{url}">
EOF
    }

    headlinks
  end
  
  def albumheader(album)
    substs = {
      'TITLE'      => album.title,
      'SHORTDESC'  => album.short_desc,
      'LONGDESC'   => album.long_desc.gsub('\[SHORT\]', album.short_desc),
      'TREE'       => album.tree,
      'ALBUMCOUNT' => album.album_count,
      'PHOTOCOUNT' => album.photo_count,
      'COUNT'      => album.count,
      'PREVLINK'   => prev_link(album),
      'NEXTLINK'   => next_link(album),
    }
    return album_substitutions(get_template('albumheader'), substs, album)
  end

  def indexentry(album)
    is_linked = album.class.name == "AlbumLink"
    substs = {
      'DEFAULTPHOTO' => defaultphoto(album),
      'TITLE'        => album.title,
      'LINK'         => is_linked ? '@' : '',
      'SHORTDESC'    => album.short_desc,
      'LONGDESC'     => album.long_desc,
      'DIRECTORY'    => url_encode(album.self_link),
      'COUNT'        => album.count,
    }
    return album_substitutions(get_template('indexentry'), substs, album)
  end

  def beginthumbs(album)
    substs = {
      'PAGECOUNT' => page_count(album),
      'PREVLINK'  => prev_link(album),
      'NEXTLINK'  => next_link(album)
    }
    return album_substitutions(get_template('beginthumbs'), substs, album)
  end

  def thumbnail(photo)
    substs = {
      'LINKS'       => proc { image_links(photo) },
      'THUMBNAIL'   => url_encode(@config.thumbnail_type.create(photo)),
      'TITLE'       => photo.title,
      'DESCRIPTION' => photo.get_field('description'),
      'BESTIMAGE'   => best_image(photo)
    } 
    return make_substitutions(get_template('thumbnail'), substs)
  end

  def albumfooter(album)
    substs = {
      'PAGELINKS' => numbered_links(album),
      'PREVLINK'  => prev_link(album),
      'NEXTLINK'  => next_link(album)
    }
    return album_substitutions(get_template('albumfooter'), substs, album)
  end

  # Private methods

  def page_count(album)
    return "&nbsp;" if album.total_pages == 1
    "#{album.current_page}/#{album.total_pages}"
  end

  def next_link(album)
    url = album.next_url
    return "&nbsp;" unless url
    return '<a href="' + url + 
           '" title="Next page">' + get_template('next') + "</a>"
  end

  def prev_link(album)
    url = album.prev_url
    return "&nbsp;" unless url
    return '<a href="' + url + \
           '" title="Previous page">' + get_template('prev') + "</a>"
  end

  def numbered_links(album)
    return "&nbsp;" if album.total_pages <= 1
    links = []
    album.total_pages.times do |i|
      if i+1 == album.current_page
        links.push((i+1).to_s)
      else
        links.push(%|<a href="#{album.index_file_url(i+1)}">#{i+1}</a>|)
      end
    end 
    return links.join("&nbsp;")
  end

  def defaultphoto(album)
    default_photo = album.default_photo
    return "&nbsp;" unless default_photo
    substs = { 
      'DIRECTORY'          => url_encode(album.self_link),
      'THUMBNAIL'          => url_encode(@config.thumbnail_type.create(default_photo)),
      'THUMBNAILDIRECTORY' => url_encode(album.default_photo_path),
      'TITLE'              => default_photo.title
    }
    return album_substitutions(get_template('defaultphoto'), substs, album)
  end

  def best_image(photo)
    best_index = @config.image_categories.length / 2
    category = find_closest_category(@config.image_categories[best_index], photo)
    if category.width.to_i < @config.get_option('min_best_width').to_i ||
           category.height.to_i < @config.get_option('min_best_height').to_i
       category = @config.image_categories[-1]
    end

    return url_encode(category.page_filename(photo.filename))
  end

  private :page_count, :next_link, :prev_link, :numbered_links,
          :defaultphoto, :best_image

end

class ImageTemplate < HTMLTemplate

  def template_name
    'image'
  end

  def photo_substitutions(text, substs = {}, photo = nil, image_category = nil)
    text = make_substitutions(text, substs)
    return text unless photo

    album = photo.parent
    global_substs = {
      'TOP'               => proc { album.top_dir },
      'UPLINK'            => proc { up_link(photo) },
      'TOPLINK'           => proc { top_link(photo) },
      'HEADLINKS'         => proc { headlinks(album, photo, image_category) },
      'THUMBNAILNEXTLINK' => proc { thumbnail_next_link(photo, image_category) },
      'THUMBNAILPREVLINK' => proc { thumbnail_prev_link(photo, image_category) }
    }
    
    return make_substitutions(text, global_substs)
  end

  def pageheader(photo, image_category)
    substs = {
      'TITLE'       => photo.title,
      'DESCRIPTION' => photo.get_field('description'),
      'ALBUM'       => photo.album_url,
      'TREE'        => photo.tree
    }
    return photo_substitutions(get_template('pageheader'), substs, photo, image_category)
  end

  def image(photo, image_category)
    fields = ''
    photo.fields.each { |field|
      case field
      when 'description'
        fields += make_substitutions(get_template('description'),
                                     { 'DESCRIPTION' => photo.get_field(field) })
      when 'order', 'title'
      else
        fields += format_field(field.capitalize, photo.get_field(field))
      end
    }

    substs = {
      'LONGLINKS'  => image_long_links(photo, image_category),
      'FILE'       => url_encode(image_category.create(photo)),
      'NEXTLINK'   => proc { next_link(photo, image_category) },
      'PREVLINK'   => proc { prev_link(photo, image_category) },
      'TITLE'      => photo.title,
      'FILENAME'   => photo.filename,
      'DIMENSIONS' => image_category.dimensions(photo),
      'FILESIZE'   => image_category.filesize(photo),
      'IMAGETYPE'  => image_category.image_type(photo.filename),
      'FIELDS'     => fields,
      'ALBUM'      => photo.album_url
    }
    return photo_substitutions(get_template('image'), substs, photo, image_category)
  end

  ## Private methods
  
  def headlinks(album, photo, image_category)
    first = if photo == album.photos[0] then nil else album.photos[0] end
    last = if photo == album.photos[-1] then nil else album.photos[-1] end

    headlink_hash = {
      'toc'   => album.tree_file_url,
      'top'   => album.top_url,
      'up'    => album.index_file_url,
      'first' => photo_url(image_category, first),
      'last'  => photo_url(image_category, last),
      'next'  => photo_url(image_category, photo.next_photo),
      'prev'  => photo_url(image_category, photo.prev_photo),
    }
    
    headlinks = ""
    headlink_hash.each { |key,value|
      headlinks += <<EOF if value
  <link rel="#{key}" href="#{value}">
EOF
    }

    headlinks
  end

  def format_field(fieldname, fieldvalue)
    return "" if fieldvalue == ""
    substs = {
      'FIELDNAME'  => fieldname,
      'FIELDVALUE' => fieldvalue
    }
    return make_substitutions(get_template('field'), substs)
  end
  
  def up_link(photo)
    up_subbed  = get_template('up').gsub(/\[TOP\]/,  photo.parent.top_dir)
    '<a href="' + photo.album_url + '" title="Up to album">' + up_subbed  + '</a>'
  end

  def top_link(photo)
    album = photo.parent
    top_subbed = get_template('top').gsub(/\[TOP\]/, album.top_dir)
    '<a href="' + album.top_page_url + '" title="Up to top">' + top_subbed + '</a>'
  end

  def next_link(photo, image_category)
    return "&nbsp;" unless photo.next_photo
    next_subbed = get_template('next').gsub(/\[TOP\]/,  photo.parent.top_dir)
    ret = '<a href="'
    ret += photo_url(image_category, photo.next_photo)
    ret += '" title="Next photo">' + get_template('next') + '</a>'
    ret
  end

  def prev_link(photo, image_category)
    return "&nbsp;" unless photo.prev_photo
    prev_subbed = get_template('prev').gsub(/\[TOP\]/,  photo.parent.top_dir)
    ret = '<a href="'
    ret += photo_url(image_category, photo.prev_photo)
    ret += '" title="Previous photo">' + get_template('prev') + '</a>'
    ret
  end

  def thumbnail_next_link(photo, image_category)
    return "&nbsp;" unless photo.next_photo
    return thumbnail(photo.next_photo, image_category)
  end
  
  def thumbnail_prev_link(photo, image_category)
    return "&nbsp;" unless photo.prev_photo
    return thumbnail(photo.prev_photo, image_category)
  end

  def thumbnail(photo, image_category)
    substs = {
      'THUMBNAIL'   => url_encode(@config.thumbnail_type.create(photo)),
      'TITLE'       => photo.title,
      'DESCRIPTION' => photo.get_field('description'),
      'BESTIMAGE'   => photo_url(image_category, photo)
    } 
    return make_substitutions(get_template('thumbnail'), substs)
  end

  def photo_url(image_category, photo)
    return nil unless photo
    url_encode(find_closest_page(image_category, photo))
  end
  
  def find_closest_page(image_category, photo)
    return image_category.page_filename(photo.filename) \
      if photo.size_possible?(image_category)
      
    found_type = find_closest_category(image_category, photo)

    return found_type.page_filename(photo.filename)
  end
    
  private :headlinks, :format_field, :top_link, :up_link, :next_link,
      :prev_link, :photo_url, :find_closest_page

end

class TreeTemplate < HTMLTemplate

  def make_substitutions(text, substs)
    text = super(text, substs)

    global_substs = {
      'HEADLINKS' => '',
    }

    return super(text, global_substs)
  end

  def template_name
    'tree'
  end

  def treeheader(album)
    substs = {
      'TITLE' => 'Album Tree',
      'TOP'   => album.top_dir
    }
    return make_substitutions(get_template('treeheader'), substs)
  end

  def tree(album)
    substs = { 'TREE' => album.tree_page_entry }
    return make_substitutions(get_template('tree'), substs)
  end

end
